import json
import random
import threading
import time
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from queue import Empty, Queue
from typing import List, Type
from uuid import UUID

from psycopg import RawCursor
from psycopg.types.json import Json
from psycopg_pool import ConnectionPool
from pydantic import BaseModel
from uuid6 import uuid7

from hyrex import constants
from hyrex.dispatcher.dispatcher import Dispatcher
from hyrex.hyrex_queue import HyrexQueue
from hyrex.schemas import (
    CronJob,
    CronJobRun,
    DequeuedTask,
    EnqueueTaskRequest,
    QueuePattern,
    TaskRun,
    TaskStatus,
    WorkflowRunRequest,
    WorkflowStatus,
)
from hyrex.sql import cron_sql, sql, workflow_sql


class PostgresDispatcher(Dispatcher):
    def __init__(self, conn_string: str, batch_size=1000, flush_interval=0.05):
        super().__init__()
        self.conn_string = conn_string
        self.pool = ConnectionPool(
            conn_string + "?keepalives=1&keepalives_idle=60&keepalives_interval=10",
            open=True,
            max_idle=5,
        )

        self.local_queue = Queue()
        self.batch_size = batch_size
        self.flush_interval = flush_interval

        self.thread = threading.Thread(target=self._batch_enqueue, daemon=True)
        self.thread.start()
        self.stopping = False

        self.register_shutdown_handlers()

    @contextmanager
    def transaction(self):
        with self.pool.connection() as conn:
            with RawCursor(conn) as cur:
                try:
                    yield cur
                except InterruptedError:
                    conn.rollback()
                    raise
            conn.commit()

    def register_app(self, app_info: dict):
        with self.transaction() as cur:
            cur.execute(sql.REGISTER_APP_INFO_SQL, [1, json.dumps(app_info)])

    def mark_success(self, task_id: UUID, result: str):
        with self.transaction() as cur:
            cur.execute(sql.MARK_TASK_SUCCESS, [task_id])
            if result:
                cur.execute(sql.SAVE_RESULT, [task_id, result])

    def mark_failed(self, task_id: UUID):
        with self.transaction() as cur:
            cur.execute(sql.MARK_TASK_FAILED, [task_id])

    def set_log_link(self, task_id: UUID, log_link: str):
        with self.transaction() as cur:
            cur.execute(sql.SET_LOG_LINK, [task_id, log_link])

    def retry_task(self, task_id: UUID, backoff_seconds: int):
        if backoff_seconds > 0:
            scheduled_start = datetime.now(timezone.utc) + timedelta(
                seconds=backoff_seconds
            )
            with self.transaction() as cur:
                cur.execute(
                    sql.CREATE_RETRY_TASK_WITH_BACKOFF,
                    [task_id, uuid7(), scheduled_start],
                )
        else:
            with self.transaction() as cur:
                cur.execute(
                    sql.CREATE_RETRY_TASK,
                    [task_id, uuid7()],
                )

    def try_to_cancel_task(self, task_id: UUID):
        with self.transaction() as cur:
            cur.execute(sql.TRY_TO_CANCEL_TASK, [task_id])

    def task_canceled(self, task_id: UUID):
        with self.transaction() as cur:
            cur.execute(sql.TASK_CANCELED, [task_id])

    def dequeue(
        self,
        executor_id: UUID,
        task_names: list[str],
        queue: str = constants.ANY_QUEUE,
        concurrency_limit: int = 0,
    ) -> DequeuedTask:
        dequeued_task = None
        with self.transaction() as cur:
            if concurrency_limit > 0:
                cur.execute(
                    sql.FETCH_TASK_WITH_CONCURRENCY,
                    [queue, concurrency_limit, executor_id, task_names],
                )
            else:
                cur.execute(sql.FETCH_TASK, [queue, executor_id, task_names])
            row = cur.fetchone()
            if row:
                (
                    task_id,
                    durable_id,
                    root_id,
                    parent_id,
                    task_name,
                    args,
                    queue,
                    priority,
                    timeout_seconds,
                    scheduled_start,
                    queued,
                    started,
                    workflow_run_id,
                    attempt_number,
                    max_retries,
                ) = row
                dequeued_task = DequeuedTask(
                    id=task_id,
                    durable_id=durable_id,
                    root_id=root_id,
                    parent_id=parent_id,
                    task_name=task_name,
                    args=args,
                    queue=queue,
                    priority=priority,
                    timeout_seconds=timeout_seconds,
                    scheduled_start=scheduled_start,
                    queued=queued,
                    started=started,
                    workflow_run_id=workflow_run_id,
                    attempt_number=attempt_number,
                    max_retries=max_retries,
                )

        return dequeued_task

    def enqueue(self, tasks: list[EnqueueTaskRequest]):
        if not tasks:
            self.logger.error("Task list is empty - cannot enqueue.")
            return
        if self.stopping:
            self.logger.warning("Task enqueued during shutdown. May not be processed.")
        for task in tasks:
            self.local_queue.put(task)

    def _batch_enqueue(self):
        tasks = []
        last_flush_time = time.monotonic()

        while True:
            # Use a longer timeout and bulk collect tasks
            try:
                # Collect as many tasks as possible with one lock acquisition
                task = self.local_queue.get(timeout=self.flush_interval)
                if task is None:  # Stop signal
                    break

                tasks.append(task)
                # Drain queue without blocking (significantly reduces lock contention)
                while len(tasks) < self.batch_size:
                    try:
                        task = self.local_queue.get_nowait()
                        if task is None:
                            break
                        tasks.append(task)
                    except Empty:
                        break

                # Check if we should flush based on time or batch size
                current_time = time.monotonic()
                if (current_time - last_flush_time >= self.flush_interval) or len(
                    tasks
                ) >= self.batch_size:
                    if tasks:
                        self._enqueue_tasks(tasks)
                        tasks = []
                    last_flush_time = current_time

            except Empty:
                # Flush on timeout if we have tasks
                if tasks:
                    self._enqueue_tasks(tasks)
                    tasks = []
                last_flush_time = time.monotonic()

    def _enqueue_tasks(self, tasks: List[EnqueueTaskRequest]):
        """
        Inserts a batch of tasks into the database.

        :param tasks: List of tasks to insert.
        """
        task_data = (
            (
                task.id,
                task.durable_id,
                task.root_id,
                task.parent_id,
                task.task_name,
                Json(task.args),
                task.queue,
                task.max_retries,
                task.priority,
                task.timeout_seconds,
                task.idempotency_key,
                task.status,
                task.workflow_run_id,
                task.workflow_dependencies,
            )
            for task in tasks
        )

        with self.transaction() as cur:
            cur.executemany(
                sql.ENQUEUE_TASK,
                task_data,
            )

    def stop(self, timeout: float = 5.0) -> bool:
        self.logger.debug("Stopping dispatcher...")
        self.stopping = True

        # Signal the batch thread to stop and wait with timeout
        self.local_queue.put(None)
        self.thread.join(timeout=timeout)

        clean_shutdown = not self.thread.is_alive()

        # Close the connection pool
        if clean_shutdown:
            self.pool.close()
        else:
            self.logger.warning(
                "Batch thread did not stop cleanly, forcing connection pool to close"
            )
            self.pool.close(timeout=1.0)

        self.logger.debug(
            "Dispatcher stopped %s.",
            "successfully" if clean_shutdown else "with timeout",
        )
        return clean_shutdown

    def get_task_status(self, task_id: UUID) -> TaskStatus:
        with self.transaction() as cur:
            cur.execute(sql.GET_TASK_STATUS, [task_id])
            result = cur.fetchone()
            if result is None:
                raise ValueError(f"Task id {task_id} not found in DB.")
            return result[0]

    def register_executor(
        self,
        executor_id: UUID,
        executor_name: str,
        queue_pattern: str,
        queues: list[HyrexQueue],
        worker_name: str,
    ):
        with self.transaction() as cur:
            cur.execute(
                sql.REGISTER_EXECUTOR,
                [executor_id, executor_name, queue_pattern, queues, worker_name],
            )

    def disconnect_executor(self, executor_id: UUID):
        with self.transaction() as cur:
            cur.execute(sql.DISCONNECT_EXECUTOR, [executor_id])

    def executor_heartbeat(self, executor_ids: list[UUID], timestamp: datetime):
        with self.transaction() as cur:
            cur.execute(sql.EXECUTOR_HEARTBEAT, [timestamp, executor_ids])

    def update_executor_stats(self, executor_id: UUID, stats: dict):
        with self.transaction() as cur:
            cur.execute(sql.UPDATE_EXECUTOR_STATS, [executor_id, stats])

    def task_heartbeat(self, task_ids: list[UUID], timestamp: datetime):
        with self.transaction() as cur:
            cur.execute(sql.TASK_HEARTBEAT, [timestamp, task_ids])

    def get_tasks_up_for_cancel(self) -> list[UUID]:
        with self.transaction() as cur:
            cur.execute(sql.GET_TASKS_UP_FOR_CANCEL)
            return [row[0] for row in cur.fetchall()]

    def mark_running_tasks_lost(self, executor_id: UUID):
        with self.transaction() as cur:
            cur.execute(sql.MARK_RUNNING_TASKS_LOST, [executor_id])

    def save_result(self, task_id: UUID, result: str):
        with self.transaction() as cur:
            cur.execute(sql.SAVE_RESULT, [task_id, result])

    # TODO: Handle durable runs, unfinished tasks.
    def get_result(self, task_id: UUID) -> dict:
        with self.transaction() as cur:
            cur.execute(sql.FETCH_RESULT, [task_id])
            row = cur.fetchone()

            if row is None:
                return None

            result = row[0]
            return result

    def get_queues_for_pattern(self, pattern: QueuePattern) -> list[str]:
        with self.transaction() as cur:
            cur.execute(sql.GET_QUEUES_FOR_PATTERN, [pattern.postgres_pattern])
            return [row[0] for row in cur.fetchall()]

    def register_task(
        self,
        task_name: str,
        arg_schema: dict,
        default_config: dict,
        cron: str = None,
        source_code: str = None,
    ):
        with self.transaction() as cur:
            cur.execute(
                sql.UPSERT_TASK,
                [
                    task_name,
                    Json(arg_schema),
                    Json(default_config),
                    cron,
                    source_code,
                ],
            )

            cron_job_name = f"ScheduledTask-{task_name}"
            if cron:
                current_id = uuid7()
                task_request = EnqueueTaskRequest(
                    id=current_id,
                    durable_id=current_id,
                    workflow_run_id=None,
                    workflow_dependencies=None,
                    root_id=current_id,
                    parent_id=None,
                    queue="TODO",
                    status=TaskStatus.queued,
                    task_name=task_name,
                    args={},
                    max_retries=0,  # TODO
                    priority=1,  # TODO
                    timeout_seconds=None,  # TODO
                    idempotency_key=None,  # TODO
                )
                insert_task_command = cron_sql.create_insert_task_cron_expression(
                    task_request
                )
                cur.execute(
                    cron_sql.CREATE_CRON_JOB_FOR_TASK,
                    [cron, insert_task_command, cron_job_name],
                )
            else:
                cur.execute(cron_sql.TURN_OFF_CRON_FOR_TASK, [cron_job_name])

    def register_workflow(
        self,
        name: str,
        source_code: str,
        workflow_dag_json: dict,
        workflow_arg_schema: Type[BaseModel] | None,
        default_config: dict,
    ):
        with self.transaction() as cur:
            cron = None
            cur.execute(
                workflow_sql.UPSERT_WORKFLOW,
                [
                    name,
                    cron,
                    source_code,
                    Json(workflow_dag_json),
                    (
                        Json(workflow_arg_schema.model_json_schema())
                        if workflow_arg_schema
                        else None
                    ),
                    Json(default_config),
                ],
            )

    def send_workflow_run(self, workflow_run_request: WorkflowRunRequest) -> UUID:
        with self.transaction() as cur:
            cur.execute(
                workflow_sql.INSERT_WORKFLOW_RUN,
                [
                    workflow_run_request.id,
                    workflow_run_request.workflow_name,
                    Json(workflow_run_request.args),
                    workflow_run_request.queue,
                    workflow_run_request.timeout_seconds,
                    workflow_run_request.idempotency_key,
                ],
            )
            result = cur.fetchall()
            if len(result) != 1:
                raise ValueError(f"Insert workflow run failed.")
            return result[0][0]

    def advance_workflow_run(self, workflow_run_id: UUID):
        self.logger.info(f"Advancing workflow run {workflow_run_id}")

        with self.transaction() as cur:
            # First query to check status
            cur.execute(
                workflow_sql.SET_WORKFLOW_RUN_STATUS_BASED_ON_TASK_RUNS,
                [workflow_run_id],
            )
            result = cur.fetchall()

            if len(result) != 1:
                self.logger.warning(
                    "Result of SET_WORKFLOW_RUN_STATUS_BASED_ON_TASK_RUNS is not one row."
                )
                return None

            workflow_status = result[0][1]  # Status is second column
            if workflow_status in (WorkflowStatus.failed, WorkflowStatus.success):
                return None

            # Second query to advance the workflow
            cur.execute(workflow_sql.ADVANCE_WORKFLOW_RUN, [workflow_run_id])
            return None

    def get_workflow_run_args(self, workflow_run_id: UUID) -> dict:
        with self.transaction() as cur:
            cur.execute(workflow_sql.GET_WORKFLOW_RUN_ARGS, [workflow_run_id])
            result = cur.fetchone()
            return result[0] if result else None

    def acquire_scheduler_lock(self, worker_name: str) -> int | None:
        lock_duration = "2 minutes"
        with self.transaction() as cur:
            cur.execute(cron_sql.ACQUIRE_SCHEDULER_LOCK, [worker_name, lock_duration])
            result = cur.fetchone()
            return result[0] if result else None

    def pull_cron_job_expressions(self) -> list[CronJob]:
        with self.transaction() as cur:
            cur.execute(cron_sql.PULL_ACTIVE_CRON_EXPRESSIONS)
            rows = cur.fetchall()
            return [
                CronJob(
                    jobid=row[0],
                    schedule=row[1],
                    command=row[2],
                    active=row[3],
                    jobname=row[4],
                    activated_at=row[5],
                    scheduled_jobs_confirmed_until=row[6],
                    should_backfill=row[7],
                )
                for row in rows
            ]

    def update_cron_job_confirmation_timestamp(self, jobid: int):
        with self.transaction() as cur:
            cur.execute(cron_sql.UPDATE_CRON_JOB_CONFIRMATION_TS, [jobid])

    def schedule_cron_job_runs(self, cron_job_runs: List[CronJobRun]) -> None:
        if not cron_job_runs:
            return

        # Check all jobs have same ID
        all_same_id = all(job.jobid == cron_job_runs[0].jobid for job in cron_job_runs)
        if not all_same_id:
            job_ids = [job.jobid for job in cron_job_runs]
            self.logger.error(f"Got jobIds {job_ids}, {cron_job_runs}")
            raise ValueError(
                "All cronJobsRuns submitted here need to have the same job id."
            )

        # Execute the SQL
        with self.transaction() as cur:
            sql_dict = cron_sql.cron_job_runs_to_sql(cron_job_runs)
            self.logger.debug(
                f"cron-scheduling: <====== Running SQL =======>:\n\n{sql_dict['sql']}\n\n{sql_dict['values']}\n\n<==== DONE =====>\n\n"
            )
            cur.execute(sql_dict["sql"], sql_dict["values"])

        # Update confirmation timestamp
        self.update_cron_job_confirmation_timestamp(cron_job_runs[0].jobid)

    def execute_queued_cron_job_run(self) -> str | None:
        with self.transaction() as cur:
            cur.execute("SELECT execute_queued_command();")
            rows = cur.fetchall()
            if not rows:
                raise ValueError("Hyrex framework error.")
            return rows[0][0]  # "executed" or "not_found"

    def register_cron_sql_query(
        self,
        cron_job_name: str,
        cron_sql_query: str,
        cron_expr: str,
        should_backfill: bool,
    ) -> None:
        """Register a new cron job for executing a SQL query on a schedule."""
        with self.transaction() as cur:
            cur.execute(
                cron_sql.CREATE_CRON_JOB_FOR_SQL_QUERY,
                [cron_expr, cron_sql_query, cron_job_name, should_backfill],
            )

    def release_scheduler_lock(self, worker_name: str) -> None:
        """Release the scheduler lock for the specified worker."""
        with self.transaction() as cur:
            cur.execute(cron_sql.RELEASE_SCHEDULER_LOCK, [worker_name])

    def get_durable_run_tasks(self, durable_id: UUID) -> list[TaskRun]:
        with self.transaction() as cur:
            cur.execute(sql.GET_TASK_RUNS_BY_DURABLE_ID, [durable_id])
            results = cur.fetchall()

            task_runs = []
            for row in results:
                (
                    task_id,
                    task_name,
                    max_retries,
                    attempt_number,
                    status,
                    queued,
                    started,
                    finished,
                    task_result,
                ) = row

                # Create the TaskRun object without the result field
                task_run = TaskRun(
                    id=task_id,
                    task_name=task_name,
                    max_retries=max_retries,
                    attempt_number=attempt_number,
                    status=status,
                    queued=queued,
                    started=started,
                    finished=finished,
                    result=task_result if task_result is not None else {},
                )
                task_runs.append(task_run)

            return task_runs

    def get_workflow_durable_runs(self, workflow_run_id: UUID) -> list[UUID]:
        with self.transaction() as cur:
            cur.execute(sql.GET_WORKFLOW_DURABLE_RUNS, [workflow_run_id])
            results = cur.fetchall()
            return [row[0] for row in results]

    def try_to_cancel_durable_run(self, durable_id: UUID):
        with self.transaction() as cur:
            cur.execute(sql.TRY_TO_CANCEL_DURABLE_RUN, [durable_id])

    def update_executor_queues(self, executor_id: UUID, queues: list[str]):
        with self.transaction() as cur:
            cur.execute(sql.UPDATE_EXECUTOR_QUEUES, [executor_id, queues])



import json
import os
import threading
import time
from datetime import datetime, timezone
from queue import Empty, Queue
from typing import Type
from uuid import UUID

import grpc
import requests
from google.protobuf.struct_pb2 import Struct
from pydantic import BaseModel

from hyrex import constants
from hyrex.dispatcher.dispatcher import Dispatcher
from hyrex.env_vars import EnvVars
from hyrex.hyrex_queue import HyrexQueue
from hyrex.proto import gateway_pb2_grpc, requests_pb2, task_pb2
from hyrex.schemas import (
    CronJob,
    CronJobRun,
    DequeuedTask,
    EnqueueTaskRequest,
    QueuePattern,
    TaskRun,
    TaskStatus,
    WorkflowRunRequest,
)

# Define the epoch zero timestamp for comparison
EPOCH_ZERO = datetime(1970, 1, 1, tzinfo=timezone.utc)


def pydantic_aware_default(obj):
    if isinstance(obj, BaseModel):
        # If the object is a Pydantic model, call model_dump()
        # to get its dictionary representation. json.dumps can handle dicts.
        return obj.model_dump()
    # If it's not a Pydantic model and json.dumps doesn't know it,
    # let the default TypeError happen.
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")


class PerformanceDispatcher(Dispatcher):
    PERFORMANCE_SERVER_HOST = os.getenv(EnvVars.PERF_SERVER_HOST, "api.hyrex.io")
    PERFORMANCE_SERVER_PORT = os.getenv(EnvVars.PERF_SERVER_PORT, "443")

    # Status mapping between Python TaskStatus enum and proto TaskStatus enum
    _PY_TO_PROTO_STATUS = {
        TaskStatus.queued: task_pb2.TaskStatus.QUEUED,
        TaskStatus.waiting: task_pb2.TaskStatus.WAITING,
        TaskStatus.running: task_pb2.TaskStatus.RUNNING,
        TaskStatus.success: task_pb2.TaskStatus.SUCCESS,
        TaskStatus.failed: task_pb2.TaskStatus.FAILED,
        TaskStatus.up_for_cancel: task_pb2.TaskStatus.UP_FOR_CANCEL,
        TaskStatus.canceled: task_pb2.TaskStatus.CANCELED,
        TaskStatus.lost: task_pb2.TaskStatus.LOST,
        TaskStatus.skipped: task_pb2.TaskStatus.SKIPPED,
    }

    # Reverse mapping for proto to Python conversion
    _PROTO_TO_PY_STATUS = {v: k for k, v in _PY_TO_PROTO_STATUS.items()}

    def __init__(self, api_key: str, conn_string: str):
        # def __init__(self, api_key: str, batch_size=100, flush_interval=0.1):
        super().__init__()

        self.api_key = api_key
        self.api_key_metadata = (("x-api-key", self.api_key),)

        server_address = (
            f"{self.PERFORMANCE_SERVER_HOST}:{self.PERFORMANCE_SERVER_PORT}"
        )

        if os.getenv(EnvVars.LOCAL_TESTING):
            self.logger.info("Testing locally.")
            self.channel = grpc.insecure_channel(server_address)
        else:
            channel_credentials = grpc.ssl_channel_credentials()
            self.channel = grpc.secure_channel(server_address, channel_credentials)
        self.gateway_stub = gateway_pb2_grpc.GatewayServiceStub(self.channel)

        # TODO: Bring these back if we switch to batching of enqueues
        # self.local_queue = Queue()
        self.running = True
        # self.batch_size = batch_size
        # self.flush_interval = flush_interval

        # self.thread = threading.Thread(target=self._batch_enqueue, daemon=True)
        # self.thread.start()

        self.register_shutdown_handlers()

    def register_app(self, app_info: dict):
        app_info_struct = Struct()
        app_info_struct.update(app_info)

        request_proto = requests_pb2.RegisterAppRequest()
        request_proto.app_info = app_info_struct

        try:
            response = self.gateway_stub.RegisterApp(
                request_proto, metadata=self.api_key_metadata
            )
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

    def enqueue(self, tasks: list[EnqueueTaskRequest]):
        for task in tasks:
            proto_task = requests_pb2.EnqueueRequest()
            proto_task.id = str(task.id)
            proto_task.durable_id = str(task.id)
            proto_task.root_id = str(task.root_id)
            if task.parent_id:
                proto_task.parent_id = str(task.parent_id)
            if task.workflow_run_id:
                proto_task.workflow_run_id = str(task.workflow_run_id)
            if task.workflow_dependencies:
                proto_task.workflow_dependencies.extend(
                    [str(dep) for dep in task.workflow_dependencies]
                )
            proto_task.task_name = task.task_name
            proto_task.queue = task.queue
            proto_task.max_retries = task.max_retries
            proto_task.priority = task.priority
            if task.timeout_seconds is not None:
                proto_task.timeout_seconds = task.timeout_seconds
            if task.idempotency_key:
                proto_task.idempotency_key = task.idempotency_key

            try:
                # json.dumps will handle dicts, lists, strings, numbers etc. directly.
                # If it encounters a Pydantic model (either as task.args itself or nested),
                # it will call our pydantic_aware_default function.
                json_string = json.dumps(task.args, default=pydantic_aware_default)
                proto_task.args = json_string.encode("utf-8")
            except TypeError as e:
                self.logger.error(
                    f"Task {task.id}: Failed to serialize args to JSON: {e}"
                )
                raise

            proto_task.status = task_pb2.TaskStatus.QUEUED

            try:
                response = self.gateway_stub.Enqueue(
                    proto_task, metadata=self.api_key_metadata
                )
            except grpc.RpcError as e:
                self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
                raise

    def dequeue(
        self,
        executor_id: UUID,
        task_names: list[str],
        queue: str = constants.ANY_QUEUE,
        concurrency_limit: int = 0,
    ) -> DequeuedTask:
        request_proto = requests_pb2.DequeueRequest()
        request_proto.queue = queue

        try:
            response = self.gateway_stub.Dequeue(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug(f"gRPC call successful, response {response}")
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

        # No task found
        # Check for task.id instead of just task because sometimes Python parses this as an instantiated empty task
        if not response.task_run.id:
            return None

        task_run = response.task_run

        # Parse args from bytes to dict
        args_dict = {}
        if task_run.args:
            try:
                args_dict = json.loads(task_run.args.decode("utf-8"))
            except (json.JSONDecodeError, UnicodeDecodeError) as e:
                self.logger.error(f"Failed to decode task args: {e}")

        # Convert protobuf timestamp to datetime
        queued_time = datetime.fromtimestamp(task_run.queued.seconds)
        started_time = datetime.fromtimestamp(task_run.started.seconds)

        scheduled_start = None
        if task_run.HasField("scheduled_start"):
            scheduled_start = datetime.fromtimestamp(task_run.scheduled_start.seconds)

        timeout_seconds = None
        if task_run.HasField("timeout_seconds"):
            timeout_seconds = task_run.timeout_seconds

        workflow_run_id = None
        if task_run.HasField("workflow_run_id"):
            workflow_run_id = UUID(task_run.workflow_run_id)

        parent_id = None
        if task_run.parent_id:
            parent_id = UUID(task_run.parent_id)

        return DequeuedTask(
            id=UUID(task_run.id),
            durable_id=UUID(task_run.durable_id),
            root_id=UUID(task_run.root_id),
            parent_id=parent_id,
            task_name=task_run.task_name,
            args=args_dict,
            queue=task_run.queue,
            priority=task_run.priority,
            timeout_seconds=timeout_seconds,
            scheduled_start=scheduled_start,
            queued=queued_time,
            started=started_time,
            workflow_run_id=workflow_run_id,
            attempt_number=task_run.attempt_number,
            max_retries=task_run.max_retries,
        )

    # def enqueue(
    #     self,
    #     tasks: list[EnqueueTaskRequest],
    # ):
    #     for task in tasks:
    #         self.local_queue.put(task)

    # def _batch_enqueue(self):
    #     tasks = []
    #     last_flush_time = time.monotonic()
    #     while self.running or not self.local_queue.empty():
    #         timeout = self.flush_interval - (time.monotonic() - last_flush_time)
    #         if timeout <= 0:
    #             # Flush if the flush interval has passed
    #             if tasks:
    #                 self._enqueue_tasks(tasks)
    #                 tasks = []
    #             last_flush_time = time.monotonic()
    #             continue

    #         try:
    #             # Wait for a task or until the timeout expires
    #             task = self.local_queue.get(timeout=timeout)
    #             tasks.append(task)
    #             if len(tasks) >= self.batch_size:
    #                 # Flush if batch size is reached
    #                 self._enqueue_tasks(tasks)
    #                 tasks = []
    #                 last_flush_time = time.monotonic()
    #         except Empty:
    #             # No task received within the timeout
    #             if tasks:
    #                 self._enqueue_tasks(tasks)
    #                 tasks = []
    #             last_flush_time = time.monotonic()

    #     # Flush any remaining tasks when stopping
    #     if tasks:
    #         self._enqueue_tasks(tasks)

    # def _enqueue_tasks(self, tasks: list[EnqueueTaskRequest]):
    #     enqueue_url = f"{self.HYREX_PLATFORM_URL}{self.ENQUEUE_TASK_PATH}"
    #     headers = {
    #         "X-API-Key": self.api_key,
    #     }

    #     task_list_json = [task.model_dump(mode="json") for task in tasks]

    #     try:
    #         response = requests.post(enqueue_url, headers=headers, json=task_list_json)
    #         if response.status_code != 200:
    #             self.logger.error(f"Error enqueuing task: {response.status_code}")
    #             self.logger.error(f"Response body: {response.text}")
    #     except requests.exceptions.RequestException as e:
    #         self.logger.error(f"Error enqueuing task via API: {str(e)}")
    #         raise RuntimeError(f"Failed to enqueue task via API: {e}")

    def stop(self):
        """
        Stops the batching process and flushes remaining tasks.
        """
        self.logger.info("Stopping dispatcher...")
        self.running = False
        self.channel.close()
        self.logger.info("Dispatcher stopped successfully!")

    def mark_success(self, task_id: UUID, result: str):
        request_proto = requests_pb2.MarkSuccessRequest()
        request_proto.task_id = str(task_id)
        if result:
            request_proto.result = result

        try:
            response = self.gateway_stub.MarkSuccess(
                request_proto, metadata=self.api_key_metadata
            )
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC MarkSuccess call failed: {e.code()} - {e.details()}"
            )
            raise

    def mark_failed(self, task_id: UUID):
        request_proto = requests_pb2.MarkFailedRequest()
        request_proto.task_id = str(task_id)

        try:
            response = self.gateway_stub.MarkFailed(
                request_proto, metadata=self.api_key_metadata
            )
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC MarkFailed call failed: {e.code()} - {e.details()}"
            )
            raise

    def retry_task(self, task_id: UUID, backoff_seconds: int):
        # TODO: Implement
        self.logger.info("WOULD HAVE RETRIED!")
        pass

    # TODO: Implement
    def try_to_cancel_task(self, task_id: UUID):
        raise NotImplementedError("Cancellation not yet implemented on Hyrex platform")

    def task_canceled(self, task_id: UUID):
        raise NotImplementedError("Cancellation not yet implemented on Hyrex platform")

    def get_task_status(self, task_id: UUID) -> TaskStatus:
        request_proto = requests_pb2.GetTaskRunStatusRequest()
        request_proto.task_id = str(task_id)

        try:
            response = self.gateway_stub.GetTaskStatus(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug(
                f"gRPC GetTaskStatus call successful, response: {response}"
            )
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC GetTaskStatus call failed: {e.code()} - {e.details()}"
            )
            raise

        return self._PROTO_TO_PY_STATUS[response.status]

    def register_executor(
        self,
        executor_id: UUID,
        executor_name: str,
        queue_pattern: str,
        queues: list[HyrexQueue],
        worker_name: str,
    ):
        request_proto = requests_pb2.RegisterExecutorRequest()
        request_proto.executor_id = str(executor_id)
        request_proto.executor_name = executor_name
        request_proto.queue_pattern = queue_pattern
        request_proto.queues.extend([queue.name for queue in queues])
        request_proto.worker_name = worker_name

        try:
            response = self.gateway_stub.RegisterExecutor(
                request_proto, metadata=self.api_key_metadata
            )
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC RegisterExecutor call failed: {e.code()} - {e.details()}"
            )
            raise

    def disconnect_executor(self, executor_id: UUID):
        pass

    def mark_running_tasks_lost(self, executor_id: UUID):
        # TODO: Implement
        pass

    def executor_heartbeat(self, executor_ids: list[UUID], timestamp: datetime):
        pass

    def update_executor_stats(self, executor_id: UUID, stats: dict):
        pass

    def task_heartbeat(self, task_ids: list[UUID], timestamp: datetime):
        # TODO: Implement
        pass

    def get_tasks_up_for_cancel(self) -> list[UUID]:
        # TODO: Implement
        return []

    def get_queues_for_pattern(self, pattern: QueuePattern) -> list[str]:
        request_proto = requests_pb2.GetQueuesRequest()
        request_proto.max_num_queues = 10000
        request_proto.pattern = pattern.glob_pattern

        try:
            response = self.gateway_stub.GetQueues(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug(f"GetQueues gRPC call successful. response: {response}")
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

        return response.queues

    def register_task(
        self,
        task_name: str,
        arg_schema: Type[BaseModel] | None,
        default_config: dict,
        cron: str = None,
        source_code: str = None,
    ):
        # The proto structure has changed to use a Task message
        request_proto = requests_pb2.RegisterTaskDefRequest()

        # Create the TaskDef message
        task_def = task_pb2.TaskDef()
        task_def.task_name = task_name

        # Handle arg_schema
        if arg_schema:
            arg_schema_struct = Struct()
            arg_schema_struct.update(arg_schema)
            task_def.arg_schema.CopyFrom(arg_schema_struct)

        # Handle default_config
        if default_config:
            default_config_struct = Struct()
            default_config_struct.update(default_config)
            task_def.default_config.CopyFrom(default_config_struct)

        # Set optional fields
        if cron:
            task_def.cron = cron

        if source_code:
            task_def.source_code = source_code

        # Set the task in the request proto
        request_proto.task_def.CopyFrom(task_def)

        try:
            response = self.gateway_stub.RegisterTaskDef(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug("gRPC RegisterTaskDef call successful")
        except grpc.RpcError as e:
            self.logger.error(
                f"gRPC RegisterTaskDef call failed: {e.code()} - {e.details()}"
            )
            raise

    def acquire_scheduler_lock(self, worker_name: str) -> int | None:
        request_proto = requests_pb2.AcquireSchedulerLockRequest()
        request_proto.worker_name = worker_name

        try:
            response = self.gateway_stub.AcquireSchedulerLock(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug(response)
            return response.lock_id
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

    def pull_cron_job_expressions(self) -> list[CronJob]:
        return []

    def update_cron_job_confirmation_timestamp(self, jobid: int):
        pass

    def schedule_cron_job_runs(self, cron_job_runs: list[CronJobRun]):
        pass

    def register_cron_sql_query(
        self,
        cron_job_name: str,
        cron_sql_query: str,
        cron_expr: str,
        should_backfill: bool,
    ) -> None:
        pass

    def execute_queued_cron_job_run(self) -> str | None:
        pass

    def release_scheduler_lock(self, worker_name: str) -> None:
        pass

    def register_workflow(
        self,
        name: str,
        source_code: str,
        workflow_dag_json: str,
        workflow_arg_schema: Type[BaseModel] | None,
        default_config: dict,
    ):
        pass

    def send_workflow_run(self, workflow_run_request: WorkflowRunRequest) -> UUID:
        pass

    def advance_workflow_run(self, workflow_run_id: UUID):
        pass

    def get_workflow_run_args(self, workflow_run_id: UUID) -> dict:
        pass

    def get_durable_run_tasks(self, durable_id: UUID) -> list[TaskRun]:
        request_proto = requests_pb2.GetDurableTaskRunsRequest()
        request_proto.durable_id = str(durable_id)

        try:
            response = self.gateway_stub.GetDurableTasks(
                request_proto, metadata=self.api_key_metadata
            )
            self.logger.debug(f"gRPC call successful, response: {response}")
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

        python_task_runs = []
        for proto_task in response.tasks:
            try:
                # Map protobuf enum name to Python StrEnum.
                # Assumes the enum names in proto match the values in Python's TaskStatus.
                # Example: If proto has TASK_STATUS_SUCCESS = 0, .Name() might return "TASK_STATUS_SUCCESS".
                # Adjust the mapping logic if needed based on your actual enum definitions.
                # If the proto enum names are exactly the Python values (e.g., "success"), this is simpler.
                # Let's assume direct mapping for now:
                status_str = task_pb2.TaskStatus.Name(
                    proto_task.status
                ).lower()  # Or adjust based on actual names
                # Handle potential prefix if needed, e.g., status_str = status_str.replace('task_status_', '')
                current_status = TaskStatus(
                    status_str
                )  # Convert string name to Python Enum

                # Convert timestamps, checking for epoch zero (default/unset) for optional Python fields
                queued_dt = proto_task.queued.ToDatetime(
                    tzinfo=timezone.utc
                )  # Store with UTC timezone
                started_dt = proto_task.started.ToDatetime(tzinfo=timezone.utc)
                finished_dt = proto_task.finished.ToDatetime(tzinfo=timezone.utc)

                queued_val = queued_dt if queued_dt > EPOCH_ZERO else None
                started_val = started_dt if started_dt > EPOCH_ZERO else None
                finished_val = finished_dt if finished_dt > EPOCH_ZERO else None

                # Parse JSON result string
                task_result = (
                    json.loads(proto_task.result) if proto_task.result else None
                )

                python_task = TaskRun(
                    id=UUID(proto_task.id),
                    task_name=proto_task.task_name,
                    max_retries=proto_task.max_retries,
                    attempt_number=proto_task.attempt_number,
                    status=current_status,
                    queued=queued_val,
                    started=started_val,
                    finished=finished_val,
                    result=task_result,
                )
                python_task_runs.append(python_task)

            except (ValueError, json.JSONDecodeError, KeyError) as e:
                # Handle potential errors during conversion (e.g., invalid UUID, bad JSON, invalid status enum)
                self.logger.error(f"Error converting TaskRun {proto_task.id}: {e}")
                # TODO: Decide how to proceed: skip this task, raise an error, etc.
                raise

        return python_task_runs

    def get_workflow_durable_runs(self, workflow_run_id: UUID) -> list[UUID]:
        pass

    def try_to_cancel_durable_run(self, durable_id: UUID):
        pass

    def update_executor_queues(self, executor_id: UUID, queues: list[str]):
        request_proto = requests_pb2.UpdateExecutorQueuesRequest()
        request_proto.executor_id = str(executor_id)
        request_proto.queues.extend(
            queues
        )  # Use extend for repeated field instead of direct assignment

        try:
            response = self.gateway_stub.UpdateExecutorQueues(
                request_proto, metadata=self.api_key_metadata
            )
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

    def save_result(self, task_id: UUID, result: str):
        request_proto = requests_pb2.SaveTaskRunResultRequest()
        request_proto.task_id = str(task_id)
        request_proto.result = result

        try:
            response = self.gateway_stub.SaveTaskResult(
                request_proto, metadata=self.api_key_metadata
            )
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

    def get_result(self, task_id: UUID) -> dict:
        request_proto = requests_pb2.GetTaskRunResultRequest()
        request_proto.task_id = str(task_id)

        try:
            response = self.gateway_stub.GetTaskResult(
                request_proto, metadata=self.api_key_metadata
            )
            return json.loads(response.result)
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise

    def set_log_link(self, task_id: UUID, log_link: str):
        request_proto = requests_pb2.SetLogLinkRequest()
        request_proto.task_id = str(task_id)
        request_proto.log_link = log_link

        try:
            response = self.gateway_stub.SetLogLink(
                request_proto, metadata=self.api_key_metadata
            )
        except grpc.RpcError as e:
            self.logger.error(f"gRPC call failed: {e.code()} - {e.details()}")
            raise



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: gateway.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'gateway.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import requests_pb2 as requests__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rgateway.proto\x12\x17hyrex.performanceserver\x1a\x0erequests.proto2\x86\x0e\n\x0eGatewayService\x12\\\n\x07\x45nqueue\x12\'.hyrex.performanceserver.EnqueueRequest\x1a(.hyrex.performanceserver.EnqueueResponse\x12\\\n\x07\x44\x65queue\x12\'.hyrex.performanceserver.DequeueRequest\x1a(.hyrex.performanceserver.DequeueResponse\x12\x62\n\tGetQueues\x12).hyrex.performanceserver.GetQueuesRequest\x1a*.hyrex.performanceserver.GetQueuesResponse\x12w\n\x10GetTaskRunStatus\x12\x30.hyrex.performanceserver.GetTaskRunStatusRequest\x1a\x31.hyrex.performanceserver.GetTaskRunStatusResponse\x12h\n\x0bMarkSuccess\x12+.hyrex.performanceserver.MarkSuccessRequest\x1a,.hyrex.performanceserver.MarkSuccessResponse\x12\x65\n\nMarkFailed\x12*.hyrex.performanceserver.MarkFailedRequest\x1a+.hyrex.performanceserver.MarkFailedResponse\x12\x65\n\nGetTaskRun\x12*.hyrex.performanceserver.GetTaskRunRequest\x1a+.hyrex.performanceserver.GetTaskRunResponse\x12}\n\x12GetDurableTaskRuns\x12\x32.hyrex.performanceserver.GetDurableTaskRunsRequest\x1a\x33.hyrex.performanceserver.GetDurableTaskRunsResponse\x12t\n\x0fRegisterTaskDef\x12/.hyrex.performanceserver.RegisterTaskDefRequest\x1a\x30.hyrex.performanceserver.RegisterTaskDefResponse\x12\x65\n\nGetTaskDef\x12*.hyrex.performanceserver.GetTaskDefRequest\x1a+.hyrex.performanceserver.GetTaskDefResponse\x12q\n\x0eGetAllTaskDefs\x12..hyrex.performanceserver.GetAllTaskDefsRequest\x1a/.hyrex.performanceserver.GetAllTaskDefsResponse\x12w\n\x10RegisterExecutor\x12\x30.hyrex.performanceserver.RegisterExecutorRequest\x1a\x31.hyrex.performanceserver.RegisterExecutorResponse\x12\x83\x01\n\x14UpdateExecutorQueues\x12\x34.hyrex.performanceserver.UpdateExecutorQueuesRequest\x1a\x35.hyrex.performanceserver.UpdateExecutorQueuesResponse\x12h\n\x0bRegisterApp\x12+.hyrex.performanceserver.RegisterAppRequest\x1a,.hyrex.performanceserver.RegisterAppResponse\x12\x65\n\nSetLogLink\x12*.hyrex.performanceserver.SetLogLinkRequest\x1a+.hyrex.performanceserver.SetLogLinkResponse\x12\x83\x01\n\x14\x41\x63quireSchedulerLock\x12\x34.hyrex.performanceserver.AcquireSchedulerLockRequest\x1a\x35.hyrex.performanceserver.AcquireSchedulerLockResponseB\x08Z\x06proto/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'gateway_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z\006proto/'
  _globals['_GATEWAYSERVICE']._serialized_start=59
  _globals['_GATEWAYSERVICE']._serialized_end=1857
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: requests.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'requests.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from . import task_pb2 as task__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0erequests.proto\x12\x17hyrex.performanceserver\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\ntask.proto\"\x99\x04\n\x0e\x45nqueueRequest\x12\n\n\x02id\x18\x01 \x01(\t\x12\x12\n\ndurable_id\x18\x02 \x01(\t\x12\x0f\n\x07root_id\x18\x03 \x01(\t\x12\x1c\n\x0fworkflow_run_id\x18\x04 \x01(\tH\x00\x88\x01\x01\x12\x1d\n\x15workflow_dependencies\x18\x05 \x03(\t\x12\x16\n\tparent_id\x18\x06 \x01(\tH\x01\x88\x01\x01\x12\x33\n\x06status\x18\x07 \x01(\x0e\x32#.hyrex.performanceserver.TaskStatus\x12\x11\n\ttask_name\x18\x08 \x01(\t\x12\x0c\n\x04\x61rgs\x18\t \x01(\x0c\x12\r\n\x05queue\x18\n \x01(\t\x12\x13\n\x0bmax_retries\x18\x0b \x01(\x05\x12\x33\n\x08priority\x18\x0c \x01(\x0e\x32!.hyrex.performanceserver.Priority\x12\x1c\n\x0ftimeout_seconds\x18\r \x01(\x05H\x02\x88\x01\x01\x12\x1c\n\x0fidempotency_key\x18\x0e \x01(\tH\x03\x88\x01\x01\x12\x38\n\x0fscheduled_start\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x04\x88\x01\x01\x42\x12\n\x10_workflow_run_idB\x0c\n\n_parent_idB\x12\n\x10_timeout_secondsB\x12\n\x10_idempotency_keyB\x12\n\x10_scheduled_start\"\x11\n\x0f\x45nqueueResponse\"4\n\x0e\x44\x65queueRequest\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t\x12\r\n\x05queue\x18\x02 \x01(\t\"E\n\x0f\x44\x65queueResponse\x12\x32\n\x08task_run\x18\x01 \x01(\x0b\x32 .hyrex.performanceserver.TaskRun\";\n\x10GetQueuesRequest\x12\x16\n\x0emax_num_queues\x18\x01 \x01(\x05\x12\x0f\n\x07pattern\x18\x02 \x01(\t\"4\n\x11GetQueuesResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0e\n\x06queues\x18\x02 \x03(\t\"$\n\x11GetTaskRunRequest\x12\x0f\n\x07task_id\x18\x01 \x01(\t\"H\n\x12GetTaskRunResponse\x12\x32\n\x08task_run\x18\x01 \x01(\x0b\x32 .hyrex.performanceserver.TaskRun\"*\n\x17GetTaskRunStatusRequest\x12\x0f\n\x07task_id\x18\x01 \x01(\t\"O\n\x18GetTaskRunStatusResponse\x12\x33\n\x06status\x18\x01 \x01(\x0e\x32#.hyrex.performanceserver.TaskStatus\"E\n\x12MarkSuccessRequest\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12\x13\n\x06result\x18\x02 \x01(\tH\x00\x88\x01\x01\x42\t\n\x07_result\"\x15\n\x13MarkSuccessResponse\"$\n\x11MarkFailedRequest\x12\x0f\n\x07task_id\x18\x01 \x01(\t\"\x14\n\x12MarkFailedResponse\"/\n\x19GetDurableTaskRunsRequest\x12\x12\n\ndurable_id\x18\x01 \x01(\t\"Q\n\x1aGetDurableTaskRunsResponse\x12\x33\n\ttask_runs\x18\x02 \x03(\x0b\x32 .hyrex.performanceserver.TaskRun\"6\n\x11SetLogLinkRequest\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12\x10\n\x08log_link\x18\x02 \x01(\t\"\x14\n\x12SetLogLinkResponse\"L\n\x16RegisterTaskDefRequest\x12\x32\n\x08task_def\x18\x01 \x01(\x0b\x32 .hyrex.performanceserver.TaskDef\"\x19\n\x17RegisterTaskDefResponse\"&\n\x11GetTaskDefRequest\x12\x11\n\ttask_name\x18\x01 \x01(\t\"H\n\x12GetTaskDefResponse\x12\x32\n\x08task_def\x18\x01 \x01(\x0b\x32 .hyrex.performanceserver.TaskDef\"\x17\n\x15GetAllTaskDefsRequest\"M\n\x16GetAllTaskDefsResponse\x12\x33\n\ttask_defs\x18\x01 \x03(\x0b\x32 .hyrex.performanceserver.TaskDef\"\x15\n\x13RegisterCronRequest\"\x16\n\x14RegisterCronResponse\"\x81\x01\n\x17RegisterExecutorRequest\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t\x12\x15\n\rexecutor_name\x18\x02 \x01(\t\x12\x15\n\rqueue_pattern\x18\x03 \x01(\t\x12\x0e\n\x06queues\x18\x04 \x03(\t\x12\x13\n\x0bworker_name\x18\x05 \x01(\t\"\x1a\n\x18RegisterExecutorResponse\"B\n\x1bUpdateExecutorQueuesRequest\x12\x13\n\x0b\x65xecutor_id\x18\x01 \x01(\t\x12\x0e\n\x06queues\x18\x02 \x03(\t\"\x1e\n\x1cUpdateExecutorQueuesResponse\"?\n\x12RegisterAppRequest\x12)\n\x08\x61pp_info\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\"\x15\n\x13RegisterAppResponse\"D\n\x1b\x41\x63quireSchedulerLockRequest\x12\x13\n\x0bworker_name\x18\x01 \x01(\t\x12\x10\n\x08\x64uration\x18\x02 \x01(\t\"@\n\x1c\x41\x63quireSchedulerLockResponse\x12\x14\n\x07lock_id\x18\x01 \x01(\x05H\x00\x88\x01\x01\x42\n\n\x08_lock_id\"\x17\n\x15TestConnectionRequest\"\x18\n\x16TestConnectionResponseB\x08Z\x06proto/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'requests_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z\006proto/'
  _globals['_ENQUEUEREQUEST']._serialized_start=119
  _globals['_ENQUEUEREQUEST']._serialized_end=656
  _globals['_ENQUEUERESPONSE']._serialized_start=658
  _globals['_ENQUEUERESPONSE']._serialized_end=675
  _globals['_DEQUEUEREQUEST']._serialized_start=677
  _globals['_DEQUEUEREQUEST']._serialized_end=729
  _globals['_DEQUEUERESPONSE']._serialized_start=731
  _globals['_DEQUEUERESPONSE']._serialized_end=800
  _globals['_GETQUEUESREQUEST']._serialized_start=802
  _globals['_GETQUEUESREQUEST']._serialized_end=861
  _globals['_GETQUEUESRESPONSE']._serialized_start=863
  _globals['_GETQUEUESRESPONSE']._serialized_end=915
  _globals['_GETTASKRUNREQUEST']._serialized_start=917
  _globals['_GETTASKRUNREQUEST']._serialized_end=953
  _globals['_GETTASKRUNRESPONSE']._serialized_start=955
  _globals['_GETTASKRUNRESPONSE']._serialized_end=1027
  _globals['_GETTASKRUNSTATUSREQUEST']._serialized_start=1029
  _globals['_GETTASKRUNSTATUSREQUEST']._serialized_end=1071
  _globals['_GETTASKRUNSTATUSRESPONSE']._serialized_start=1073
  _globals['_GETTASKRUNSTATUSRESPONSE']._serialized_end=1152
  _globals['_MARKSUCCESSREQUEST']._serialized_start=1154
  _globals['_MARKSUCCESSREQUEST']._serialized_end=1223
  _globals['_MARKSUCCESSRESPONSE']._serialized_start=1225
  _globals['_MARKSUCCESSRESPONSE']._serialized_end=1246
  _globals['_MARKFAILEDREQUEST']._serialized_start=1248
  _globals['_MARKFAILEDREQUEST']._serialized_end=1284
  _globals['_MARKFAILEDRESPONSE']._serialized_start=1286
  _globals['_MARKFAILEDRESPONSE']._serialized_end=1306
  _globals['_GETDURABLETASKRUNSREQUEST']._serialized_start=1308
  _globals['_GETDURABLETASKRUNSREQUEST']._serialized_end=1355
  _globals['_GETDURABLETASKRUNSRESPONSE']._serialized_start=1357
  _globals['_GETDURABLETASKRUNSRESPONSE']._serialized_end=1438
  _globals['_SETLOGLINKREQUEST']._serialized_start=1440
  _globals['_SETLOGLINKREQUEST']._serialized_end=1494
  _globals['_SETLOGLINKRESPONSE']._serialized_start=1496
  _globals['_SETLOGLINKRESPONSE']._serialized_end=1516
  _globals['_REGISTERTASKDEFREQUEST']._serialized_start=1518
  _globals['_REGISTERTASKDEFREQUEST']._serialized_end=1594
  _globals['_REGISTERTASKDEFRESPONSE']._serialized_start=1596
  _globals['_REGISTERTASKDEFRESPONSE']._serialized_end=1621
  _globals['_GETTASKDEFREQUEST']._serialized_start=1623
  _globals['_GETTASKDEFREQUEST']._serialized_end=1661
  _globals['_GETTASKDEFRESPONSE']._serialized_start=1663
  _globals['_GETTASKDEFRESPONSE']._serialized_end=1735
  _globals['_GETALLTASKDEFSREQUEST']._serialized_start=1737
  _globals['_GETALLTASKDEFSREQUEST']._serialized_end=1760
  _globals['_GETALLTASKDEFSRESPONSE']._serialized_start=1762
  _globals['_GETALLTASKDEFSRESPONSE']._serialized_end=1839
  _globals['_REGISTERCRONREQUEST']._serialized_start=1841
  _globals['_REGISTERCRONREQUEST']._serialized_end=1862
  _globals['_REGISTERCRONRESPONSE']._serialized_start=1864
  _globals['_REGISTERCRONRESPONSE']._serialized_end=1886
  _globals['_REGISTEREXECUTORREQUEST']._serialized_start=1889
  _globals['_REGISTEREXECUTORREQUEST']._serialized_end=2018
  _globals['_REGISTEREXECUTORRESPONSE']._serialized_start=2020
  _globals['_REGISTEREXECUTORRESPONSE']._serialized_end=2046
  _globals['_UPDATEEXECUTORQUEUESREQUEST']._serialized_start=2048
  _globals['_UPDATEEXECUTORQUEUESREQUEST']._serialized_end=2114
  _globals['_UPDATEEXECUTORQUEUESRESPONSE']._serialized_start=2116
  _globals['_UPDATEEXECUTORQUEUESRESPONSE']._serialized_end=2146
  _globals['_REGISTERAPPREQUEST']._serialized_start=2148
  _globals['_REGISTERAPPREQUEST']._serialized_end=2211
  _globals['_REGISTERAPPRESPONSE']._serialized_start=2213
  _globals['_REGISTERAPPRESPONSE']._serialized_end=2234
  _globals['_ACQUIRESCHEDULERLOCKREQUEST']._serialized_start=2236
  _globals['_ACQUIRESCHEDULERLOCKREQUEST']._serialized_end=2304
  _globals['_ACQUIRESCHEDULERLOCKRESPONSE']._serialized_start=2306
  _globals['_ACQUIRESCHEDULERLOCKRESPONSE']._serialized_end=2370
  _globals['_TESTCONNECTIONREQUEST']._serialized_start=2372
  _globals['_TESTCONNECTIONREQUEST']._serialized_end=2395
  _globals['_TESTCONNECTIONRESPONSE']._serialized_start=2397
  _globals['_TESTCONNECTIONRESPONSE']._serialized_end=2421
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import requests_pb2 as requests__pb2

GRPC_GENERATED_VERSION = '1.71.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in gateway_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class GatewayServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Enqueue = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/Enqueue',
                request_serializer=requests__pb2.EnqueueRequest.SerializeToString,
                response_deserializer=requests__pb2.EnqueueResponse.FromString,
                _registered_method=True)
        self.Dequeue = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/Dequeue',
                request_serializer=requests__pb2.DequeueRequest.SerializeToString,
                response_deserializer=requests__pb2.DequeueResponse.FromString,
                _registered_method=True)
        self.GetQueues = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetQueues',
                request_serializer=requests__pb2.GetQueuesRequest.SerializeToString,
                response_deserializer=requests__pb2.GetQueuesResponse.FromString,
                _registered_method=True)
        self.GetTaskRunStatus = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetTaskRunStatus',
                request_serializer=requests__pb2.GetTaskRunStatusRequest.SerializeToString,
                response_deserializer=requests__pb2.GetTaskRunStatusResponse.FromString,
                _registered_method=True)
        self.MarkSuccess = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/MarkSuccess',
                request_serializer=requests__pb2.MarkSuccessRequest.SerializeToString,
                response_deserializer=requests__pb2.MarkSuccessResponse.FromString,
                _registered_method=True)
        self.MarkFailed = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/MarkFailed',
                request_serializer=requests__pb2.MarkFailedRequest.SerializeToString,
                response_deserializer=requests__pb2.MarkFailedResponse.FromString,
                _registered_method=True)
        self.GetTaskRun = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetTaskRun',
                request_serializer=requests__pb2.GetTaskRunRequest.SerializeToString,
                response_deserializer=requests__pb2.GetTaskRunResponse.FromString,
                _registered_method=True)
        self.GetDurableTaskRuns = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetDurableTaskRuns',
                request_serializer=requests__pb2.GetDurableTaskRunsRequest.SerializeToString,
                response_deserializer=requests__pb2.GetDurableTaskRunsResponse.FromString,
                _registered_method=True)
        self.RegisterTaskDef = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/RegisterTaskDef',
                request_serializer=requests__pb2.RegisterTaskDefRequest.SerializeToString,
                response_deserializer=requests__pb2.RegisterTaskDefResponse.FromString,
                _registered_method=True)
        self.GetTaskDef = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetTaskDef',
                request_serializer=requests__pb2.GetTaskDefRequest.SerializeToString,
                response_deserializer=requests__pb2.GetTaskDefResponse.FromString,
                _registered_method=True)
        self.GetAllTaskDefs = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/GetAllTaskDefs',
                request_serializer=requests__pb2.GetAllTaskDefsRequest.SerializeToString,
                response_deserializer=requests__pb2.GetAllTaskDefsResponse.FromString,
                _registered_method=True)
        self.RegisterExecutor = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/RegisterExecutor',
                request_serializer=requests__pb2.RegisterExecutorRequest.SerializeToString,
                response_deserializer=requests__pb2.RegisterExecutorResponse.FromString,
                _registered_method=True)
        self.UpdateExecutorQueues = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/UpdateExecutorQueues',
                request_serializer=requests__pb2.UpdateExecutorQueuesRequest.SerializeToString,
                response_deserializer=requests__pb2.UpdateExecutorQueuesResponse.FromString,
                _registered_method=True)
        self.RegisterApp = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/RegisterApp',
                request_serializer=requests__pb2.RegisterAppRequest.SerializeToString,
                response_deserializer=requests__pb2.RegisterAppResponse.FromString,
                _registered_method=True)
        self.SetLogLink = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/SetLogLink',
                request_serializer=requests__pb2.SetLogLinkRequest.SerializeToString,
                response_deserializer=requests__pb2.SetLogLinkResponse.FromString,
                _registered_method=True)
        self.AcquireSchedulerLock = channel.unary_unary(
                '/hyrex.performanceserver.GatewayService/AcquireSchedulerLock',
                request_serializer=requests__pb2.AcquireSchedulerLockRequest.SerializeToString,
                response_deserializer=requests__pb2.AcquireSchedulerLockResponse.FromString,
                _registered_method=True)


class GatewayServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Enqueue(self, request, context):
        """---- Valkey requests ----
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Dequeue(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetQueues(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTaskRunStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def MarkSuccess(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def MarkFailed(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTaskRun(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetDurableTaskRuns(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RegisterTaskDef(self, request, context):
        """---- Postgres requests ----
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTaskDef(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAllTaskDefs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RegisterExecutor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateExecutorQueues(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RegisterApp(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetLogLink(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AcquireSchedulerLock(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GatewayServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Enqueue': grpc.unary_unary_rpc_method_handler(
                    servicer.Enqueue,
                    request_deserializer=requests__pb2.EnqueueRequest.FromString,
                    response_serializer=requests__pb2.EnqueueResponse.SerializeToString,
            ),
            'Dequeue': grpc.unary_unary_rpc_method_handler(
                    servicer.Dequeue,
                    request_deserializer=requests__pb2.DequeueRequest.FromString,
                    response_serializer=requests__pb2.DequeueResponse.SerializeToString,
            ),
            'GetQueues': grpc.unary_unary_rpc_method_handler(
                    servicer.GetQueues,
                    request_deserializer=requests__pb2.GetQueuesRequest.FromString,
                    response_serializer=requests__pb2.GetQueuesResponse.SerializeToString,
            ),
            'GetTaskRunStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTaskRunStatus,
                    request_deserializer=requests__pb2.GetTaskRunStatusRequest.FromString,
                    response_serializer=requests__pb2.GetTaskRunStatusResponse.SerializeToString,
            ),
            'MarkSuccess': grpc.unary_unary_rpc_method_handler(
                    servicer.MarkSuccess,
                    request_deserializer=requests__pb2.MarkSuccessRequest.FromString,
                    response_serializer=requests__pb2.MarkSuccessResponse.SerializeToString,
            ),
            'MarkFailed': grpc.unary_unary_rpc_method_handler(
                    servicer.MarkFailed,
                    request_deserializer=requests__pb2.MarkFailedRequest.FromString,
                    response_serializer=requests__pb2.MarkFailedResponse.SerializeToString,
            ),
            'GetTaskRun': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTaskRun,
                    request_deserializer=requests__pb2.GetTaskRunRequest.FromString,
                    response_serializer=requests__pb2.GetTaskRunResponse.SerializeToString,
            ),
            'GetDurableTaskRuns': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDurableTaskRuns,
                    request_deserializer=requests__pb2.GetDurableTaskRunsRequest.FromString,
                    response_serializer=requests__pb2.GetDurableTaskRunsResponse.SerializeToString,
            ),
            'RegisterTaskDef': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterTaskDef,
                    request_deserializer=requests__pb2.RegisterTaskDefRequest.FromString,
                    response_serializer=requests__pb2.RegisterTaskDefResponse.SerializeToString,
            ),
            'GetTaskDef': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTaskDef,
                    request_deserializer=requests__pb2.GetTaskDefRequest.FromString,
                    response_serializer=requests__pb2.GetTaskDefResponse.SerializeToString,
            ),
            'GetAllTaskDefs': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAllTaskDefs,
                    request_deserializer=requests__pb2.GetAllTaskDefsRequest.FromString,
                    response_serializer=requests__pb2.GetAllTaskDefsResponse.SerializeToString,
            ),
            'RegisterExecutor': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterExecutor,
                    request_deserializer=requests__pb2.RegisterExecutorRequest.FromString,
                    response_serializer=requests__pb2.RegisterExecutorResponse.SerializeToString,
            ),
            'UpdateExecutorQueues': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateExecutorQueues,
                    request_deserializer=requests__pb2.UpdateExecutorQueuesRequest.FromString,
                    response_serializer=requests__pb2.UpdateExecutorQueuesResponse.SerializeToString,
            ),
            'RegisterApp': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterApp,
                    request_deserializer=requests__pb2.RegisterAppRequest.FromString,
                    response_serializer=requests__pb2.RegisterAppResponse.SerializeToString,
            ),
            'SetLogLink': grpc.unary_unary_rpc_method_handler(
                    servicer.SetLogLink,
                    request_deserializer=requests__pb2.SetLogLinkRequest.FromString,
                    response_serializer=requests__pb2.SetLogLinkResponse.SerializeToString,
            ),
            'AcquireSchedulerLock': grpc.unary_unary_rpc_method_handler(
                    servicer.AcquireSchedulerLock,
                    request_deserializer=requests__pb2.AcquireSchedulerLockRequest.FromString,
                    response_serializer=requests__pb2.AcquireSchedulerLockResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'hyrex.performanceserver.GatewayService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('hyrex.performanceserver.GatewayService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class GatewayService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Enqueue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/Enqueue',
            requests__pb2.EnqueueRequest.SerializeToString,
            requests__pb2.EnqueueResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Dequeue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/Dequeue',
            requests__pb2.DequeueRequest.SerializeToString,
            requests__pb2.DequeueResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetQueues(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetQueues',
            requests__pb2.GetQueuesRequest.SerializeToString,
            requests__pb2.GetQueuesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTaskRunStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetTaskRunStatus',
            requests__pb2.GetTaskRunStatusRequest.SerializeToString,
            requests__pb2.GetTaskRunStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def MarkSuccess(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/MarkSuccess',
            requests__pb2.MarkSuccessRequest.SerializeToString,
            requests__pb2.MarkSuccessResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def MarkFailed(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/MarkFailed',
            requests__pb2.MarkFailedRequest.SerializeToString,
            requests__pb2.MarkFailedResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTaskRun(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetTaskRun',
            requests__pb2.GetTaskRunRequest.SerializeToString,
            requests__pb2.GetTaskRunResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetDurableTaskRuns(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetDurableTaskRuns',
            requests__pb2.GetDurableTaskRunsRequest.SerializeToString,
            requests__pb2.GetDurableTaskRunsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RegisterTaskDef(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/RegisterTaskDef',
            requests__pb2.RegisterTaskDefRequest.SerializeToString,
            requests__pb2.RegisterTaskDefResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTaskDef(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetTaskDef',
            requests__pb2.GetTaskDefRequest.SerializeToString,
            requests__pb2.GetTaskDefResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetAllTaskDefs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/GetAllTaskDefs',
            requests__pb2.GetAllTaskDefsRequest.SerializeToString,
            requests__pb2.GetAllTaskDefsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RegisterExecutor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/RegisterExecutor',
            requests__pb2.RegisterExecutorRequest.SerializeToString,
            requests__pb2.RegisterExecutorResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UpdateExecutorQueues(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/UpdateExecutorQueues',
            requests__pb2.UpdateExecutorQueuesRequest.SerializeToString,
            requests__pb2.UpdateExecutorQueuesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RegisterApp(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/RegisterApp',
            requests__pb2.RegisterAppRequest.SerializeToString,
            requests__pb2.RegisterAppResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SetLogLink(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/SetLogLink',
            requests__pb2.SetLogLinkRequest.SerializeToString,
            requests__pb2.SetLogLinkResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AcquireSchedulerLock(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/hyrex.performanceserver.GatewayService/AcquireSchedulerLock',
            requests__pb2.AcquireSchedulerLockRequest.SerializeToString,
            requests__pb2.AcquireSchedulerLockResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from psycopg.types.json import Json
from psycopg_pool import ConnectionPool

from hyrex.dispatcher.postgres_dispatcher import PostgresDispatcher
from hyrex.hyrex_queue import HyrexQueue
from hyrex.schemas import EnqueueTaskRequest
from hyrex.sql import cron_sql, sql


# Single-threaded variant of Postgres dispatcher. (Slower enqueuing.)
class PostgresLiteDispatcher(PostgresDispatcher):
    def __init__(self, conn_string: str):
        self.conn_string = conn_string
        self.pool = ConnectionPool(
            conn_string + "?keepalives=1&keepalives_idle=60&keepalives_interval=10",
            open=True,
            max_idle=5,
        )

        self.register_shutdown_handlers()

    def enqueue(
        self,
        tasks: list[EnqueueTaskRequest],
    ):
        task_data = (
            (
                task.id,
                task.durable_id,
                task.root_id,
                task.parent_id,
                task.task_name,
                Json(task.args),
                task.queue,
                task.max_retries,
                task.priority,
                task.timeout_seconds,
                task.idempotency_key,
                task.status,
                task.workflow_run_id,
                task.workflow_dependencies,
            )
            for task in tasks
        )

        with self.transaction() as cur:
            cur.executemany(
                sql.ENQUEUE_TASK,
                task_data,
            )

    def stop(self):
        """
        Stops the batching process and flushes remaining tasks.
        """
        self.logger.debug("Stopping dispatcher...")
        self.pool.close()
        self.logger.debug("Dispatcher stopped successfully!")
